'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.useAsyncQueryDoc = exports.useBatchQueryDoc = exports.useQueryDoc = exports.useAsyncQueryIds = exports.useBatchQueryIds = exports.useQueryIds = exports.emit = undefined;

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

exports.useModel = useModel;
exports.useOn = useOn;
exports.useEmit = useEmit;
exports.generateUseQueryIds = generateUseQueryIds;
exports.generateUseQueryDoc = generateUseQueryDoc;
exports.useLocalDoc = useLocalDoc;
exports.useSession = useSession;
exports.usePage = usePage;

var _react = require('react');

var _model = require('@startupjs/model');

var _model2 = _interopRequireDefault(_model);

var _types = require('./types');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var emit = exports.emit = _model2.default.emit.bind(_model2.default);

function useModel() {
  for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return (0, _react.useMemo)(function () {
    return _model2.default.scope.apply(_model2.default, args);
  }, [].concat(args));
}

function useOn() {
  for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    args[_key2] = arguments[_key2];
  }

  (0, _react.useLayoutEffect)(function () {
    var eventName = args[0];

    var listener = _model2.default.on.apply(_model2.default, args);
    return function () {
      _model2.default.removeListener(eventName, listener);
    };
  });
}

function useEmit() {
  return emit;
}

function generateUseQueryIds() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      batch = _ref.batch,
      optional = _ref.optional;

  var useFn = batch ? _types.useBatchQuery : optional ? _types.useAsyncQuery : _types.useQuery;
  return function (collection) {
    var ids = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    var _useFn = useFn(collection, { _id: { $in: ids } }),
        _useFn2 = (0, _slicedToArray3.default)(_useFn, 3),
        $items = _useFn2[1],
        ready = _useFn2[2];

    if (!ready) return [undefined, $items, ready];
    if (options.reverse) ids = ids.slice().reverse();
    var items = ids.map(function (id) {
      return _model2.default.get(collection + '.' + id);
    }).filter(Boolean);
    return [items, $items, ready];
  };
}

var useQueryIds = exports.useQueryIds = generateUseQueryIds();
var useBatchQueryIds = exports.useBatchQueryIds = generateUseQueryIds({ batch: true });
var useAsyncQueryIds = exports.useAsyncQueryIds = generateUseQueryIds({ optional: true });

function generateUseQueryDoc() {
  var _ref2 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      batch = _ref2.batch,
      optional = _ref2.optional;

  var useFn = batch ? _types.useBatchQuery : optional ? _types.useAsyncQuery : _types.useQuery;
  return function (collection, query) {
    query = (0, _extends3.default)({}, query, {
      $limit: 1
    });
    if (!query.$sort) query.$sort = { createdAt: -1 };

    var _useFn3 = useFn(collection, query),
        _useFn4 = (0, _slicedToArray3.default)(_useFn3, 3),
        _useFn4$ = _useFn4[0],
        items = _useFn4$ === undefined ? [] : _useFn4$,
        ready = _useFn4[2];

    var itemId = items[0] && items[0].id;
    var $item = (0, _react.useMemo)(function () {
      if (!itemId) return;
      return _model2.default.at(collection + '.' + itemId);
    }, [itemId]);
    if (!ready || !itemId) return [undefined, undefined, ready];
    return [_model2.default.get(collection + '.' + itemId), $item, ready];
  };
}

var useQueryDoc = exports.useQueryDoc = generateUseQueryDoc();
var useBatchQueryDoc = exports.useBatchQueryDoc = generateUseQueryDoc({ batch: true });
var useAsyncQueryDoc = exports.useAsyncQueryDoc = generateUseQueryDoc({ optional: true });

function useLocalDoc(collection, docId) {
  console.warn('\n    useLocalDoc() is DEPRECATED! Instead use useDoc() the same exact way.\n    useLocalDoc() will be removed in the next release!\n  ');
  if (typeof collection !== 'string') {
    throw new Error('[react-sharedb] useLocalDoc(): `collection` must be a String. Got: ' + collection);
  }
  if (!docId) {
    console.warn('\n      [react-sharedb] useLocalDoc(): You are trying to subscribe to an undefined document id:\n        ' + collection + '.' + docId + '\n      Falling back to \'__NULL__\' document to prevent critical crash.\n      You should prevent situations when the `docId` is undefined.\n    ');
    docId = '__NULL__';
  }
  return (0, _types.useLocal)(collection + '.' + docId);
}

function useSession(path) {
  if (typeof path !== 'string') {
    throw new Error('[react-sharedb] useSession(): `path` must be a String. Got: ' + path);
  }
  return (0, _types.useLocal)('_session' + '.' + path);
}

function usePage(path) {
  if (typeof path !== 'string') {
    throw new Error('[react-sharedb] usePage(): `path` must be a String. Got: ' + path);
  }
  return (0, _types.useLocal)('_page' + '.' + path);
}